<?php
if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

/**
 * WorthstockSync è dove sono presenti i metodi corrispondenti a ogni tipo di chiamata delle API WORTHSTOCK, implementando la logica che assembla le informazioni necessarie ed effettua la richiesta
 * Contiene anche i metodi a basso livello CURL per stabilire la comunicazione
 */
class WorthstockSync extends WorthstockBase {

    const API_LIVE_URL = "https://www.worthstock.com";
    //const API_TEST_URL = "https://ws04-worthstock-test.valeositi.it";
    const API_TEST_URL = "https://sandbox.worthstock.com";
    const CACHE_DIR = "worthstock";
    private $api_url;

    /* public function __construct()
    {
        if ( $this->main()->settings->get_option( 'worthstock_test_environment_flag' ) == 'yes' ) {
            $this->api_url = self::API_TEST_URL;
        } else {
            $this->api_url = self::API_LIVE_URL;
        }
    } */

    protected function get_api_url() {
        if ( !$this->api_url ) {
            if ( $this->main()->settings->get_option( 'worthstock_test_environment_flag' ) == 'yes' ) {
                $this->api_url = self::API_TEST_URL;
            } else {
                $this->api_url = self::API_LIVE_URL;
            }
        }
        return $this->api_url;
    }

	private static function get_user_agent() {
		return isset( $_SERVER['HTTP_USER_AGENT'] ) ? $_SERVER['HTTP_USER_AGENT'] : null;
	}

	private static function get_referer() {
		return isset( $_SERVER['HTTP_REFERER'] ) ? $_SERVER['HTTP_REFERER'] : null;
	}

	// return a comma-separated list of role names for the given user
	public static function get_user_roles( $user_id ) {
		$roles = false;

		if ( !class_exists('WP_User') )
			return false;

		if ( $user_id > 0 ) {
			$comment_user = new WP_User( $user_id );
			if ( isset( $comment_user->roles ) )
				$roles = join( ',', $comment_user->roles );
		}

		if ( is_multisite() && is_super_admin( $user_id ) ) {
			if ( empty( $roles ) ) {
				$roles = 'super_admin';
			} else {
				$comment_user->roles[] = 'super_admin';
				$roles = join( ',', $comment_user->roles );
			}
		}

		return $roles;
	}


	/**
	 * Log debugging info to the error log.
	 *
	 * Enabled when WP_DEBUG_LOG is enabled (and WP_DEBUG, since according to
	 * core, "WP_DEBUG_DISPLAY and WP_DEBUG_LOG perform no function unless
	 * WP_DEBUG is true)
	 *
	 * @param mixed $worthstock_debug The data to log.
	 */
	public static function log( $worthstock_debug ) {
		if ( apply_filters( 'worthstock_debug_log', defined( 'WP_DEBUG' ) && WP_DEBUG && defined( 'WP_DEBUG_LOG' ) && WP_DEBUG_LOG && defined( 'WORTHSTOCK__DEBUG' ) && WORTHSTOCK__DEBUG ) ) {
			error_log( print_r( compact( 'worthstock_debug' ), true ) );
		}
	}

    /**
     * output to logger (using woocommerce logger)
     */
    function api_logger($message, $context_name='worthstock', $type='info') {
        if ( !$message OR !function_exists('wc_get_logger') ) {return false;}

        $logger = wc_get_logger();
        
        $context = array( 'source' => $context_name );

        if (is_array($message)) {$message = json_encode($message);}
            
        $logger->info( $message, $context );
        
    }

    /**
     * check status from api response
     * @param object $response
     * @return null|boolean
     */
    function check_response_status($response) {
        if ( !isset($response->status) || empty($response->status) ) return null;
        if (strtoupper($response->status) == "SUCCESS") {
            return true;
        } else {
            return false;
        }
    }

    function log_status_operations($operation_name, $response, $product_id=null) {
        if ( !isset($response->status) || empty($response->status) ) return;
        $subject = ($product_id ? " (".$product_id.")" : null);
        
        $this->save_product_last_operation($operation_name, $response->status, $product_id);
        if (strtoupper($response->status) == "SUCCESS") {
            $message = $operation_name . $subject . " ha avuto successo";
        } else {
            $message = $operation_name . $subject . " non ha avuto successo";
        }
        $this->api_logger($message);
        //if (WORTHSTOCK__DEBUG === true) {            
        if ($this->main()->settings->get_option( 'worthstock_debug_flag' ) == 'yes' ) {
            $this->api_logger('RESPONSE ' . json_encode($response));
        }
    }

    function save_product_last_operation($operation_name, $operation_status, $product_id) {
        if ($product_id) {
            update_post_meta( $product_id, '_worthstock_last_operation', array($operation_name, time(), strtolower($operation_status)) );
        }
    }
    
    function read_product_last_operation($product_id) {
        return get_post_meta($product_id, '_worthstock_last_operation', true);
    }

	public function api_login($exposed=false, $pars=[]) {
        $setWebhookUrl = isset($pars['setWebhookUrl']) ? $pars['setWebhookUrl']: null;
        if ( isset($pars['ApiUrl']) && !empty($pars['ApiUrl']) ) {
            $this->api_url = $pars['ApiUrl'];
        }
        $url = $this->getEndPoint("api_login");
        // if missing credentials get them
        if (!$pars['Username'] || !$pars['Password']) {
            $pars = $this->getLoginCredential();
        }
        // if still missing credentials, abort
        if (!$pars['Username'] || !$pars['Password']) {
            return false;
        }
        if ($setWebhookUrl === false) {
            $pars['WebhookUrl'] = "";
        }
        $response = $this->curlCaller($url, true, $pars);
        if (strtoupper($response->status) == "SUCCESS") {            
            $token = $response->results->Keys[0];
            $this->cacheToken($token);
            $message = "Autenticazione effettuata con successo";
            $this->api_logger($message);
            // se exposed chiamata rest esterna
            if ($exposed) {
                print (json_encode(["success" => true,"token"=>$token, "data" => array("message" => $message)]));
                wp_die();
            }
            return $token;
        } else {
            $message = "Autenticazione non riuscita";
            $this->api_logger($message);
            // se exposed chiamata rest esterna
            if ($exposed) {
                print (json_encode(["success" => false, "data" => array("message" => $message)]));
                wp_die();
            }
            return false;
        }
    }

    /**
     * wrapper di api_login() per ottenere token di auth dalla versione in cache se presente
     */
    public function getToken() {
        $token = get_transient('_worthstock_token');
        //TODO add hashing
        if ( ! $token ) {
            $token = $this->api_login();
        }
        return $token;
    }

    private function cacheToken($token) {
        $token = set_transient('_worthstock_token', $token, HOUR_IN_SECONDS);        
    }

    public function api_get_single_product($token,$product) {
        $this->main()->set_woocommerce_replication(false);
	    $url=$this->getEndPoint("api_get_single_product");
        $url.="?IdProduct=$product";
        $response = $this->getGetData($url,$token);
        $product_array=$response->results["Products"];
        
        $this->createUpdateSingleProduct( (array) $product_array[0]);
        //var_dump($product_array);
        $message = "Prodotto ricevuto e salvato";        
        $this->api_logger($message);
        print (json_encode(["success" => true, "data" => array("message" => $message, "sample_data" => $product_array)]));
        wp_die();
    }

    public function api_get_all_products($token) {
        $this->main()->set_woocommerce_replication(false);
        $url = $this->getEndPoint("api_get_all_products_published");
        //if ( ! $response = $this->readJsonFile(__FUNCTION__) ) {
            $response = $this->getGetData($url, $token);
            $this->saveJsonResponse(__FUNCTION__, json_encode($response));
        //}
        $product_array= (array)$response->results->Products;
        $count=0;
        foreach ($product_array as $product){
            $product_data=(array)$product;
            //var_dump($product_data);
            //echo '<br>';
            if($count <= 10) {
                $this->createUpdateSingleProduct($product_data);
            }
            $count++;
        }
        $message = "Prodotti ricevuti e salvati";        
        $this->api_logger($message);
        print (json_encode(["success" => true,"token"=>$token, "data" => array("message" => $message, "sample_data" => $product_array[0])]));
        wp_die();
    }

    public function api_get_all_products_slim($IdProducts=[]) {
        $url = $this->getEndPoint("api_get_all_products_slim");
        $token = $this->getToken();;
        if (is_array($IdProducts) && count($IdProducts)) {
            $data = array("IdProducts" => $IdProducts);
        }
        return $response = wp_remote_get($url, array(
            'body'=> ($data) ? wp_json_encode($data) : '',
            'headers' => array(
                'Authorization' => ' Bearer '.$token,
            )
        ));
    }    

    public function api_get_all_products_drafted($IdProducts=[]) {
        $url = $this->getEndPoint("api_get_all_products");
        $token = $this->getToken();
        if (is_array($IdProducts) && count($IdProducts)) {
            //$data = array("IdProduct" => $IdProducts);
            $url .= '?IdProduct=' . join(',',$IdProducts);
        }
        $response = wp_remote_get($url, array(
            //'body'=> ($data) ? wp_json_encode($data) : '',
            'headers' => array(
                'Authorization' => ' Bearer '.$token,
            )
        ));
        // Retrieve information
        $response_code = wp_remote_retrieve_response_code($response);
        $response_message = wp_remote_retrieve_response_message($response);
        $response_body = wp_remote_retrieve_body($response);
        $response_body = json_decode($response_body,true);
        if ($response_code == 200 && $response_body['status'] == 'SUCCESS') {
            return $response_body['results'];
        }
        
        /* if ( !is_wp_error($response) ) {
            return new WP_REST_Response(
                array(
                    'status' => $response_code,
                    'response' => $response_message,
                    'body_response' => $response_body
                )
            );
        } else {
            return new WP_Error($response_code, $response_message, $response_body);
        }
        exit; */
        return false;
    }

    public function api_get_categories($token) {
        $this->main()->set_woocommerce_replication(false);
        $url=$this->getEndPoint("api_get_categories")."?Language=IT";        
        // if ( ! $response = $this->readJsonFile(__FUNCTION__) ) {
        $response = $this->getGetData($url,$token);
        $this->saveJsonResponse(__FUNCTION__, json_encode($response));
        // }
        $category_data=$response->results->Categories;
        foreach($category_data as $cat_obj){
            $cat_data_array=(array)$cat_obj;
            $this->addCategories($cat_data_array);
        }
        $message = "Categorie ricevute e salvate";        
        $this->api_logger($message);
        print (json_encode(["success" => true, "data" => array("message" => $message, "sample_data" => $category_data[0])]));
        wp_die();
    }

    /**
     * resolve remote endpoint by action
     */
    private function getEndPoint($action){
        switch ($action) {
            case "api_login":
                return $this->get_api_url()."/api/v1/Login";
                break;
            case "api_get_single_product":
                return $this->get_api_url()."/api/v1/Products/Get";
            case "api_get_single_product_published":
                return $this->get_api_url()."/api/v1/Products/Get/Published";
            case "api_get_all_products_slim":
                return $this->get_api_url()."/api/v1/Products/GetProductsSlim";
            case "api_get_all_products":
                return $this->get_api_url()."/api/v1/Products/Get";                
            case "api_put_product":
            case "api_post_product":
            case "api_delete_product":
                return $this->get_api_url()."/api/v1/Products";
            case "api_get_all_products_published":
                return $this->get_api_url()."/api/v1/Products/Get/Published";                
            case "api_get_categories":
                return $this->get_api_url()."/api/v1/Categories";
            case "api_put_prices":
                return $this->get_api_url()."/api/v1/Variants/Prices";
            case "api_put_quantity":
                return  $this->get_api_url()."/api/v1/Variants/Quantity";
        }
    }


    protected function getApiWoocommerceUrl() {
        return $this->main()->receiver->returnUpdateProductStockEndpoint();
    }

    /**
     * retrieve API credentials
     */
    private function getLoginCredential() {
        // TODO LOGIN INTERATTIVO SENZA SALVARE PASSWORD IN DB
        $woocommerce_api_url = $this->getApiWoocommerceUrl();
        if ( $this->main()->settings->get_option( 'worthstock_test_environment_flag' ) == 'yes' ) {
            $login = $this->main()->settings->get_option( 'api_test_login' );
            $password = $this->main()->settings->get_option( 'api_test_secret' );
        } else {
            $login = $this->main()->settings->get_option( 'api_login' );
            $password = $this->main()->settings->get_option( 'api_secret' );
        }
        $defaults = array(
            'WebhookUrl'=> $woocommerce_api_url
        );
        return wp_parse_args(['Username' => $login, 'Password' => $password], $defaults);
    }

    protected function getProductBySku( $sku, $post_type='product' ) {
        global $wpdb;
        $results = $wpdb->get_var( "
            SELECT p.ID FROM {$wpdb->prefix}posts as p
            JOIN {$wpdb->prefix}postmeta as pm ON p.ID = pm.post_id
            WHERE p.post_type LIKE '".$post_type."' AND p.post_status LIKE 'publish'
            AND meta_key = '_sku' AND meta_value = '$sku' LIMIT 1;" );
        if( count($results) > 0 ) {
            return $results;
        } else {
            return false;
        }
    }

    /**
     * trova prodotto woocommerce by IdProduct
     */
    public function getProductByWorthstockId( $id_product, $post_type='product' ) {
        global $wpdb;
        if ($post_type == 'product') {
            $product_id = $wpdb->get_var( $wpdb->prepare( "SELECT post_id, meta_key, meta_value, wp.* FROM `wp_postmeta` wpm LEFT JOIN wp_posts wp ON wpm.post_id = wp.ID WHERE wp.post_type= 'product' AND wpm.meta_key='_wst_id_product' AND meta_value='%s' LIMIT 1", $id_product ) );
        } else {
            $product_id = $wpdb->get_var( $wpdb->prepare( "SELECT post_id FROM $wpdb->postmeta WHERE meta_key='_wst_id_product' AND meta_value='%s' LIMIT 1", $id_product ) );
        }
        return $product_id;
    }

    protected function getVariationByShopifyId($_unique_id_shopify) {
        global $wpdb;
        $variation_id = $wpdb->get_var( $wpdb->prepare( "SELECT post_id FROM $wpdb->postmeta WHERE meta_key='_wst_unique_id_shopify' AND meta_value='%s' LIMIT 1", $_unique_id_shopify ) );
        return $variation_id;
    }

    
    public function getVariantByWorthstockId($_id_product_variant) {
        global $wpdb;
        $variation_id = $wpdb->get_var( $wpdb->prepare( "SELECT post_id FROM $wpdb->postmeta WHERE meta_key='_wst_id_product_variant' AND meta_value='%s' LIMIT 1", $_id_product_variant ) );
        return $variation_id;
    }
    
    public function countProductsByWorthstockId( $post_type='product' ) {
        global $wpdb;
        $synced_products = $wpdb->get_var( $wpdb->prepare( "SELECT COUNT(pm.post_id) FROM $wpdb->posts as p LEFT JOIN $wpdb->postmeta as pm ON p.ID = pm.post_id WHERE meta_key='%s' AND p.post_type = '%s'", '_wst_id_product', 'product') );

        /* $r = $wpdb->get_results(  $wpdb->prepare( "
        SELECT pm.meta_value AS name, count(*) AS count  FROM {$wpdb->postmeta} pm
        LEFT JOIN {$wpdb->posts} p ON p.ID = pm.post_id
        WHERE pm.meta_key = '%s'
        AND pm.meta_value != '' 
        AND p.post_type = '%s'
        GROUP BY pm.meta_value
        ORDER BY pm.meta_value          
        ", $key, $type) 
        );
    return $r; */
        return $synced_products;
    }

    public static function getAllProductInSync( $post_type='product' ) {
        global $wp;
        $metakey = WorthstockDataMapper::getWorthstockCustomParamName('id_product');
        $args = array(
            'posts_per_page' => -1,
            'post_type' => $post_type,
            'meta_query' => array(
                'relation' => 'AND',
                array(
                    'key' => 'some_key',
                    'compare' => 'EXISTS',
                ),
                array(
                    'key' => $metakey,
                    'compare' => '!=',
                    'value' => ''
                )
            )
        );
        $posts = get_posts($args);
        return $posts;
    }

    public static function getAllMetaValues($metakey) {        
        global $wpdb;
        return $wpdb->get_col("SELECT DISTINCT meta_value
        FROM $wpdb->postmeta WHERE meta_key = '$metakey'" );
    }

    protected function getLanguageData($product_data,$language="IT"){
	    $language_data=$product_data["Languages"];
	    foreach ($language_data as $lang){
	        $lang=(array)$lang;
	        if(strtoupper($lang["Language"])==strtoupper($language)){
	            return $lang;
            }
        }
	    return false;
    }

    protected function createProduct($product_data){

    }

    protected function createUpdateSingleProduct($product_data) {	    
        // Starting from Sku, check if product already exist. if not,we create a new one
        $product_language_data= $this->getLanguageData($product_data, "IT");
        $post_id = $this->getProductBySku($product_language_data["ProductUrl"]);
        if($post_id == false) {
            $post_args = array(
                'post_title' => sanitize_text_field($product_language_data["ProductName"]), // The product's Title
                'post_type' => 'product',
                'post_status' => 'publish' // This could also be $data['status'];
            );
            $post_id = wp_insert_post( $post_args );
        // ora mi faccio restituire direttamente il valore scalare
        // }else{
        //     $post_id= $post_id[0];
        }
        // If the post exist or was created okay, let's try update the WooCommerce values.
        if ( ! empty( $post_id ) ) {
            //$product = wc_get_product( $post_id );                
            $product = new WC_Product_Variable( $post_id );
            $product->set_sku( $product_language_data["ProductUrl"] ); // Generate a SKU with a prefix. (i.e. 'pre-123')
            //$product->set_regular_price( $product_data[""] );// Be sure to use the correct decimal price.
            //$product->set_sale_price( $product_data[""] );// Be sure to use the correct decimal price.
            //$product->set_category_ids( array( 16, 17 ) ); // Set multiple category ID's.
            // conversione KG/gr
            if ( isset($product_data["Weight"]) && ($product_data["Weight"]) && ($weight = $product_data["Weight"]/1000) )
                $product->set_weight($weight);
            // GrossWeight non presente in woocommerce
            // if ( isset($product_data["GrossWeight"]) && ($product_data["GrossWeight"]) && ($weight = $product_data["GrossWeight"]/1000) )
            //     $product->set_weight($weight);
            if ( isset($product_data["Width"]) && ($product_data["Width"]) )
                $product->set_length($product_data["Width"]);//Width
            if ( isset($product_data["Height"]) && ($product_data["Height"]) )
                $product->set_height($product_data["Height"]);//Height            
            if ( isset($product_data["Depth"]) && ($product_data["Depth"]) )
                $product->set_width($product_data["Depth"]); //Depth
            $product->set_description($product_language_data["Description"]);
            $product->set_short_description($product_language_data["Details"]);
            // $this->getProdCatByApiId($product_data["IdCategory"]);
            //$product_data["IdCategory"] = '1605887047';
            if ( $cat_id = $this->getProdCatByApiId($product_data["IdCategory"]) ) {
                $product->set_category_ids([$cat_id]);
            }
            $product->save(); // Save/update the WooCommerce order object.
        }
        $params=$this->main()->mapper->getWorthstockCustomParamList();
        foreach($params as $key=>$value){
            update_post_meta( $post_id, $value, $product_data[$key] );
        }
        $this->getSingleProductImg($product_data["Variants"],$post_id);
        //$this->create_product_attribute( $post_id, $product_data["Variants"] );
        $this->createVariants($product_data["Variants"],$post_id);
        wc_delete_product_transients( $post_id );
        $this->save_product_last_operation('Prodotto importato in Woocommerce', 'success', $post_id);
        //echo "--- ID CATEGORY:.".$product_data["IdCategory"]." <br>";
        //$this->getCategoryByApiId($product_data["IdCategory"]);
        //var_dump($zippo);

    }

    function upload_images_to_product($product_id, $image_id_array) {
        //take the first image in the array and set that as the featured image
        set_post_thumbnail($product_id, $image_id_array[0]);

        //if there is more than 1 image - add the rest to product gallery
        if(sizeof($image_id_array) > 1) {
            array_shift($image_id_array); //removes first item of the array (because it's been set as the featured image already)
            update_post_meta($product_id, '_product_image_gallery', implode(',',$image_id_array)); //set the images id's left over after the array shift as the gallery images
        }
    }

    public function getSingleProductImg($variants,$product_id){
        foreach($variants as $variant){
            $variant = (array) $variant;
            $galleryIDs = [];
            foreach($variant["Images"] as $index => $image) {
                // avoid the useless image-not-available placeholder. TODO: this break thumbnail in categories page
                if ( !strstr($image, 'image-not-available') ) {
                    $attachmentId = $this->UploadImgToMediaFromUrl($image, $product_id);
                    if ($variant['DefaultVariant'] == 1) {
                        set_post_thumbnail($product_id, $attachmentId);
                        //break;
                    }
                    $galleryIDs[] = $attachmentId;
                }
            }
            if ( !empty ($galleryIDs)) {
                update_post_meta($product_id, '_product_image_gallery', implode(',',$galleryIDs));
            }
        }
    }

    /**
     * Get attachment_id by image filename.
     *
     * @param string $img The image name with extension after dot.
     *
     * @return bool|int
     */
    function checkImageByFilename( $img ) {
        global $wpdb;
        $img = '%/' . $img;
        $sql = $wpdb->prepare(
            "SELECT post_id FROM $wpdb->postmeta WHERE meta_key = '_wp_attached_file' AND meta_value LIKE '%s'",
            $img
        );
        return $wpdb->get_var( $sql ) !== null ? $wpdb->get_var( $sql ) : false;
    } 

    /**
     * associa i meta di worthstock ricevuti in response al prodotto woo
     */
    function save_worthstock_custom_product_meta($post_id, $products_data, $action_type='update') {
        $products_data=(array) (property_exists($products_data, 'Products')) ? $products_data->Products : $products_data;
        if (empty($products_data)) return;

        $product = wc_get_product( $post_id );
        // salvo per futuri controlli
        $this->main()->queue->save_serialized_product($post_id);

        // first save the products ids
        delete_post_meta($post_id, $this->main()->mapper->getWorthstockCustomParamName('id_product'));
        foreach($products_data as $index => $product_data) {
            if ($product_data->IdProduct) {
                add_post_meta( $post_id, $this->main()->mapper->getWorthstockCustomParamName('id_product'), $product_data->IdProduct);
            }
        }

        foreach($products_data as $index => $product_data) {
            /* if (is_array($variations)) {
                reset($variations);
            } */
            
            $product_data=(array) $product_data;

            if ($index === 0) {
                $params = $this->main()->mapper->getWorthstockCustomParamList();
                foreach($params as $key => $value) {
                    if ( isset($product_data[$key]) && !empty($product_data[$key]) ) {
                        if ($key != "IdProduct") {
                            // non salvo più altri meta su update, per via di differenze pubblicato e bozze
                            if ($action_type == 'insert') {
                                update_post_meta( $post_id, $this->main()->mapper->getWorthstockCustomParamName($value), $product_data[$key] );
                            }
                        }
                    }
                }
            }

            /* if ($product_data['IdProduct']) {
                add_post_meta( $post_id, $this->main()->mapper->getWorthstockCustomParamName('id_product'), $product_data['IdProduct']);
            } */
              
            $variants = $product_data['Variants'];
            if ( is_array($variants) && count($variants) ) {
                foreach ($variants as $variant) {                
                    $variant = (array) $variant;
                    //$variation_id = $this->getVariantByWorthstockId($variant['IdVariant']);
                    $variation_id = $this->getVariationByShopifyId($variant['UniqueIdShopify']);
                    
                    if ($variation_id) {
                        update_post_meta( $variation_id, $this->main()->mapper->getWorthstockCustomParamName('id_product_variant'), $variant["IdVariant"] );
                        update_post_meta( $variation_id, $this->main()->mapper->getWorthstockCustomParamName('id_product'), $product_data["IdProduct"] );
                    } else if ( $product->is_type('variable') === false ) {
                        update_post_meta( $product->get_id(), $this->main()->mapper->getWorthstockCustomParamName('id_product_variant'), $variant["IdVariant"]);                  
                    }
                }
            }
            
        }
    }

    /**
     * prepare and map variant data
     */
    protected function createVariants( $variants, $product_id ) {
        if (!$product_id) return;
        $product = wc_get_product( $product_id );
        
        $variants_data = [];
        $variants_attributes = [];
        // filtro sul tipo attributo TODO: rimuovere
        //$allowed_attributes = array("Size", "Color", "Material");
        $allowed_attributes = $this->main()->mapper->default_attributes;
        if ( is_array($variants) && count($variants) ) {
            foreach ($variants as $variant) {
                $variant = (array) $variant; // cast from object to array
                $variants_data = [];
                // vediamo se esiste variante
                $variant_sku = $product->get_sku() . "-" . $variant['IdVariant'];
                $variant_id = $this->getProductBySku($variant_sku, 'product_variation');
                if ($variant_id) {
                    continue;
                } else {
                    $variants_data['sku'] = $variant_sku;
                }
                $variants_data['_wst_id_product_variant'] = $variant["IdVariant"];
                // TODO save extra prices as custom_meta?
                $variants_data['regular_price'] = $variant["PriceB2C"];
                $variants_data['sale_price'] = $variant["PricePROMO"];
                $variants_data['stock_qty'] = $variant["QuantityAvailable"];
                // foreach($variant as $key => $value) {
                //     if ( in_array($key, $allowed_attributes) ) {
                //         $variants_attributes[$key] = $value;
                //     }
                // }
                foreach($variant["Languages"] as $language) {
                    //TODO ??? sbagliato
                    if ($language->Language != 'IT') continue; // prendiamo in esame solo IT
                        
                    foreach($language as $attribute => $term) {
                        if ( in_array($attribute, $allowed_attributes) ) {
                            $variants_attributes[$attribute] = $term;
                        }
                    }
                }
                $variants_data["attributes"] = $variants_attributes;                
                $this->create_product_variation( $product_id, $variants_data );
            }
        }
        /*
        [IdVariant] => 5
        [QuantityAvailable] => 0
        [DefaultVariant] => 1
        [PriceLST] => 140.00
        [PriceB2C] => 110.00
        [Images] => Array (URLs)
        [Languages] => Array

        $variation_data['attributes']
        $variation_data['sku']
        $variation_data['regular_price']
        $variation_data['sale_price']
        $variation_data['stock_qty']
        $variation_data['images'] ?        
        */
    }
    
    public function list_default_product_attributes( $variants ) {
        $attributes = [];
        //$allowed_attributes = array("Size", "Color", "Material");
        $allowed_attributes = $this->main()->mapper->default_attributes;
        if ( is_array($variants) && count($variants) ) {
            foreach ($variants as $variant) {
                $variant = (array) $variant; // cast from object to array
                foreach($variant["Languages"] as $language) {
                    if ($language->Language != 'IT') continue; // prendiamo in esame solo IT
                        
                    foreach($language as $attribute => $term) {
                        if ( in_array($attribute, $allowed_attributes) ) {
                            $attributes[$attribute][] = $term;
                        }
                    }
                }               
            }
        }
        return $attributes;
    }

    function create_product_attribute( $product_id, $variants ) {
        $attributes = $this->list_default_product_attributes( $variants );
        foreach($attributes as $variation_attribute => $values) {
            $variation_attribute = mb_strtolower($variation_attribute);

            //Create main product
            //$product = new WC_Product_Variable($product_id);
            $product = wc_get_product( $product_id ); 
    
            //Create the attribute object
            $attribute = new WC_Product_Attribute();
    
            //pa_size tax id
            $attribute->set_id( 1 ); // -> SET to 0
    
            //pa_size slug
            $attribute->set_name( $variation_attribute ); // -> removed 'pa_' prefix
    
            //Set terms slugs
            $attribute->set_options(
                // array(
                //     'blue',
                //     'grey'
                // )
                $values
            );
    
            $attribute->set_position( 0 );
    
            //If enabled
            $attribute->set_visible( 1 );
    
            //If we are going to use attribute in order to generate variations
            $attribute->set_variation( 1 );        

            $product->set_attributes(array($attribute));
    
            //Save main product to get its id
            $id = $product->save();

        }
    }

    function alt_create_product_variation( $product_id, $variation_data ) {

        $variation = new WC_Product_Variation();
        $variation->set_parent_id($product_id);

        //Set attributes requires a key/value containing
        // tax and term slug
        $variation->set_attributes(
            // array(
            //     'size' => 'blue' // -> removed 'pa_' prefix
            // )
            $variation_data["attributes"]
        );
        ## Set/save all other data

        // SKU
        if( ! empty( $variation_data['sku'] ) )
            $variation->set_sku( $variation_data['sku'] );

        // Prices
        if( empty( $variation_data['sale_price'] ) ){
            $variation->set_price( $variation_data['regular_price'] );
        } else {
            $variation->set_price( $variation_data['sale_price'] );
            $variation->set_sale_price( $variation_data['sale_price'] );
        }
        $variation->set_regular_price( $variation_data['regular_price'] );

        // Stock
        if( ! empty($variation_data['stock_qty']) ){
            $variation->set_stock_quantity( $variation_data['stock_qty'] );
            $variation->set_manage_stock(true);
            $variation->set_stock_status('');
        } else {
            $variation->set_manage_stock(false);
        }

        $variation->set_weight(''); // weight (reseting)
        $variation->save(); // Save the data
    }

    function create_product_variation( $product_id, $variation_data ){
        // Get the Variable product object (parent)
        // $product = wc_get_product($product_id);

        // $variation_post = array(
        //     'post_title'  => $product->get_name(),
        //     'post_name'   => 'product-'.$product_id.'-variation',
        //     'post_status' => 'publish',
        //     'post_parent' => $product_id,
        //     'post_type'   => 'product_variation',
        //     'guid'        => $product->get_permalink()
        // );
        // // Creating the product variation
        // $variation_id = wp_insert_post( $variation_post );


        // Get an instance of the WC_Product_Variation object
        // $variation = new WC_Product_Variation( $variation_id );

            
        $global_attributes = wc_get_attribute_taxonomies();

        $global_attributes_slugs = wp_list_pluck( $global_attributes, 'attribute_name' );
        
        $variation = new WC_Product_Variation();
        $variation->set_parent_id($product_id);
        
        // SKU
        if( ! empty( $variation_data['sku'] ) )
            $variation->set_sku( $variation_data['sku'] );

            
        // Stock
        if( ! empty($variation_data['stock_qty']) ){
            $variation->set_stock_quantity( $variation_data['stock_qty'] );
            $variation->set_manage_stock(true);
            $variation->set_stock_status('');
        } else {
            $variation->set_manage_stock(false);
        }
        
        
        $variation->set_status('publish');   
        $variation->set_attributes(
            // array(
            //     'size' => 'blue' // -> removed 'pa_' prefix
            // )
            $variation_data["attributes"]
        );
        $variation->save();
        $variation_id = $variation->get_id();        

        // id_variation
        update_post_meta( $variation_id, $this->main()->mapper->getWorthstockCustomParamName('id_product_variant'), $variation_data['_wst_id_product_variant'] );

        // Iterating through the variations attributes
        foreach ($variation_data['attributes'] as $attribute => $term_name )
        {
            $attribute = mb_strtolower($attribute);
            $taxonomy = 'pa_'.$attribute; // The attribute taxonomy

            // If taxonomy doesn't exists we create it (Thanks to Carl F. Corneil)
            if ( ! taxonomy_exists( $taxonomy ) ) {
                register_taxonomy(
                    $taxonomy,
                    'product_variation',
                    array (
                        'hierarchical' => false,
                        'label' => ucfirst( $attribute ),
                        'query_var' => true,
                        'rewrite' => array( 'slug' => sanitize_title($attribute) ), // The base slug
                    )
                );
            }

            // crea la tassonomia attributo globale in modo che sia attivata a ogni esecuzione
            if ( ! in_array( $attribute, $global_attributes_slugs ) ) {

                $args = array(
                    'slug'    => $attribute,
                    'name'   => __( ucfirst($attribute), 'worthstock' ),
                    'type'    => 'select',
                    'orderby' => 'menu_order',
                    'has_archives'  => false,
                );

                $result = wc_create_attribute( $args );
            }

            // Check if the Term name exist and if not we create it.
            if( ! term_exists( $term_name, $taxonomy ) )
                wp_insert_term( $term_name, $taxonomy ); // Create the term

            $term_slug = get_term_by('name', $term_name, $taxonomy )->slug; // Get the term slug
            /* 
            // Get the post Terms names from the parent variable product.
            $post_term_names =  wp_get_post_terms( $product_id, $taxonomy, array('fields' => 'names') );

            // Check if the post term exist and if not we set it in the parent variable product.
            if( ! in_array( $term_name, $post_term_names ) )
                wp_set_post_terms( $product_id, $term_name, $taxonomy, true ); */

            // Set/save the attribute data in the product variation
            update_post_meta( $variation_id, 'attribute_'.$taxonomy, $term_slug );
         
        }

        ## Set/save all other data

        // SKU
        if( ! empty( $variation_data['sku'] ) )
            $variation->set_sku( $variation_data['sku'] );

        // Prices
        if( empty( $variation_data['sale_price'] ) ){
            $variation->set_price( $variation_data['regular_price'] );
        } else {
            $variation->set_price( $variation_data['sale_price'] );
            $variation->set_sale_price( $variation_data['sale_price'] );
        }
        $variation->set_regular_price( $variation_data['regular_price'] );

        // Stock
        if( ! empty($variation_data['stock_qty']) ){
            $variation->set_stock_quantity( $variation_data['stock_qty'] );
            $variation->set_manage_stock(true);
            $variation->set_stock_status('');
        } else {
            $variation->set_manage_stock(false);
        }

        $variation->set_weight(''); // weight (reseting)
        $variation->save(); // Save the data
    }

    protected function addCategories($category_data, $level=0) {
        $name=$category_data["Title"];
        $description=$category_data["Description"];
        $category_id=$category_data["Id"];
        $image = $this->get_api_url().$category_data["ImageUrlTopBanner"];
        $slug=sanitize_title($name);
        $id=term_exists($slug,'product_cat');
        // if Category exists
        if($id) {
            //get existing category id
            $category = get_term_by( 'slug', $slug, 'product_cat' );
            $cat_id = $category->term_id;
        } else {
            // create new category
            $cat_id = $this->AddCategory($name,$description,$level);
        }
        // Saving Metadatas
        update_term_meta($cat_id,"idcategory",$category_id);
        update_term_meta($cat_id,"image_url_top_banner",$category_data["ImageUrlTopBanner"]);
        update_term_meta($cat_id,"image_url_top_banner_small",$category_data["ImageUrlTopBannerSmall"]);
        update_term_meta($cat_id,"image_url_menu_banner",$category_data["ImageUrlMenuBanner"]);
        // avoid the useless image-not-available placeholder. TODO: this break thumbnail in categories page
        if ( !strstr($image, 'image-not-available') ) {
            $attachmentId=$this->UploadImgToMediaFromUrl($image);

            update_woocommerce_term_meta( $cat_id, 'thumbnail_id', $attachmentId );
        }
        // subcategories
        if ($category_data["Nodes"]) {
            //$level++;
            foreach($category_data["Nodes"] as $cat_obj){
                $cat_data_array=(array)$cat_obj;
                $this->addCategories($cat_data_array, $cat_id);
            }
        }
    }

    /**
     * callback che è stata definita, ma non so con quale intenzione: la creo vuota intanto
     */
    public function save_pos_cat_custom_meta( $category_id ) {
		// global $wpdb;

		// if ( ! isset( $this->edited_product_cats[ $category_id ] ) ) {
		// 	return;
		// }

		// $category_object = get_term( $category_id, 'product_cat' );
		// $prev_parent     = $this->edited_product_cats[ $category_id ];
		// $new_parent      = $category_object->parent;

		// // No edits - no need to modify relationships.
		// if ( $prev_parent === $new_parent ) {
		// 	return;
		// }

		// $this->delete( $category_id, $prev_parent );
		// $this->update( $category_id );
	}

    /* callback che è stata definita, ma non so con quale intenzione: la creo vuota intanto */
    public function pos_new_cat_meta_field( $category_id ) {
    }

    public function pos_edit_cat_meta_field( $term ) {
		$idcategory = get_term_meta( $term->term_id, 'idcategory', true );
		$image_url_top_banner = get_term_meta( $term->term_id, 'image_url_top_banner', true );
		$image_url_top_banner_small = get_term_meta( $term->term_id, 'image_url_top_banner_small', true );
		$image_url_menu_banner = get_term_meta( $term->term_id, 'image_url_menu_banner', true );
        if ($idcategory) {
            echo "<p>Worthstock ID: " . $idcategory . "</p>";
        }
        if ($image_url_top_banner_small) {
            echo "<p>image_url_top_banner_small: " . $image_url_top_banner_small . "</p>";
        }
        if ($image_url_menu_banner) {
            echo "<p>image_url_menu_banner: " . $image_url_menu_banner . "</p>";
        }
        if ($image_url_top_banner) {
            echo "<p>image_url_top_banner: " . $image_url_top_banner . "</p>";
        }

    }
    
    
    /**
     * save cached copied of json response
     */
    protected function saveJsonResponse($caller, $response) {
        $upload_dir = wp_upload_dir();
        $upload_dir = $upload_dir['basedir'] . "/". self::CACHE_DIR ."/";
        if ( !file_exists($upload_dir) ) {
            mkdir($upload_dir, 0777, true);
        }
        $path = $upload_dir.$caller.'.json';
        file_put_contents($path, $response);
    }

    /**
     * read json from cache
     */
    protected function readJsonFile($caller) {
        $upload_dir = wp_upload_dir();
        $path = $upload_dir['basedir']."/".$caller.'.json';
        if (file_exists($path)) { // more checks here
            // Get the contents of the JSON file 
            $strJsonFileContents = file_get_contents($path);
            // Convert to array 
            return json_decode($strJsonFileContents);
        }
        return false;
    }
    
    protected function saveImg($url,$name){
        $upload_dir = wp_upload_dir();
        $path = $upload_dir['basedir']."/".$name;
        file_put_contents($path, file_get_contents($url));
        return $path;
    }

    // protected function AddImgToMediaFromUrl($url){
    //     require_once ABSPATH . 'wp-admin/includes/file.php';
    //     require_once ABSPATH . 'wp-admin/includes/media.php';
    //     require_once ABSPATH . 'wp-admin/includes/image.php';
    //     $desc = '';
    //     $file = $url;
    //     $file_array  = [ 'name' => wp_basename( $file ).".jpeg", 'tmp_name' => download_url( $file ) ];

    //     // If error storing temporarily, return the error.
    //     if ( is_wp_error( $file_array['tmp_name'] ) ) {
    //         return $file_array['tmp_name'];
    //     }

    //     // Do the validation and storage stuff.
    //     $id = media_handle_sideload( $file_array, 0, $desc );
    //     return $id;
    // }

    function UploadImgToMediaFromUrl($url, $product_id=0) {
        require_once ABSPATH . 'wp-admin/includes/file.php';
        require_once ABSPATH . 'wp-admin/includes/media.php';
        require_once ABSPATH . 'wp-admin/includes/image.php';
        $image = "";
        if($url != "") {
            $filename = wp_basename( $url ).".jpeg";
            if ( $attachmentId = $this->checkImageByFilename($filename) ) {
                return $attachmentId;
            }
         
            $file = array();
            //$file['name'] = $url.".jpeg";
            //$file['name'] =  $product_id.".jpeg";
            $file['name'] = $filename;
            $file['tmp_name'] = download_url($url);
            //$file['type'] = 'image/jpeg';
            $file['type'] = mime_content_type($file['tmp_name']);
            $file['size'] = filesize($file['tmp_name']);
     
            if (is_wp_error($file['tmp_name'])) {
                @unlink($file['tmp_name']);
                var_dump( $file['tmp_name']->get_error_messages( ) );
            } else {
                $attachmentId = media_handle_sideload($file, $product_id);
                 
                if ( is_wp_error($attachmentId) ) {
                    @unlink($file['tmp_name']);
                    var_dump( $attachmentId->get_error_messages( ) );
                } else {                
                    $image = wp_get_attachment_url( $attachmentId );
                }
            }
        }
        return $attachmentId;
    }

    // protected function getCategoryByApiId($api_id){
    //     $child_categories = get_categories(
    //         array(
    //             'meta_query' => array(
    //                 array(
    //                     'key'     => '$category_id',     // Adjust to your needs!
    //                     'value'   => $api_id,   // Adjust to your needs!
    //                     'compare' => '=',         // Default
    //                 )
    //             )
    //         )
    //     );
    //     //var_dump($child_categories);

    // }

    public function getProdCatByApiId($api_id){
        $args = array(
            'hide_empty' => false, // also retrieve terms which are not used yet
            'meta_query' => array(
                array(
                    'key'       => 'idcategory',
                    'value'     => $api_id,
                    'compare'   => '='
                )
            ),
            'taxonomy'  => 'product_cat',
        );
        $terms = get_terms( $args );
        if ($terms) {
            return $terms[0]->term_id;
        }
        return false;
    }

    public function getProdCatApiIdById($term_id) {
        //Get Taxonomy Meta
        return get_term_meta($term_id,'idcategory'); 
    }

    protected function AddCategory($name,$description,$parent_id){
        $cat=wp_insert_term( $name, 'product_cat', array(
            'description' => $description, // optional
            'parent' => $parent_id, // optional
            'slug' => sanitize_title($name) // optional
        ) );
        return $cat["term_id"];
    }


    function wh_taxonomy_add_new_meta_field() {
        ?>
        <div class="form-field">
            <label for="wh_meta_title"><?php _e('IdCategory', 'worthstock'); ?></label>
            <input type="text" name="wh_id_category" id="wh_id_category">
            <p class="description"><?php _e('Category ID', 'worthstock'); ?></p>
        </div>
        <?php
    }

    function wh_taxonomy_edit_meta_field($term) {
        //getting term ID
        $term_id = $term->term_id;
        // retrieve the existing value(s) for this meta field.
        $wh_id_category = get_term_meta($term_id, 'wh_id_category', true);
        ?>
        <tr class="form-field">
            <th scope="row" valign="top"><label for="wh_id_category"><?php _e('Category ID', 'worthstock'); ?></label></th>
            <td>
                <input type="text" name="wh_id_category" id="wh_id_category" value="<?php echo esc_attr($wh_id_category) ? esc_attr($wh_id_category) : ''; ?>">
                <p class="description"><?php _e('Category ID ', 'worthstock'); ?></p>
            </td>
        </tr>
        <?php
    }

    function wh_save_taxonomy_custom_meta($term_id) {
        $wh_meta_title = filter_input(INPUT_POST, 'wh_id_category');
        update_term_meta($term_id, 'idcategory', $wh_meta_title);
    }



    protected function getImg($image_url,$image_name){
        require_once('wp-admin/includes/image.php');
        require_once('wp-admin/includes/file.php');
        require_once('wp-admin/includes/media.php');
        $media = media_sideload_image($image_url,0);
        $attachments = get_posts(array(
            'post_type' => 'attachment',
            'post_status' => null,
            'post_parent' => 0,
            'orderby' => 'post_date',
            'order' => 'DESC'
        ));
        return $attachments[0]->ID;
    }


    /**
     * UNINSTALL
     */    

    public function delete_remote_products($remove_all=false) {
        if ( $remove_all !== true ) return;
        $metakey = WorthstockDataMapper::getWorthstockCustomParamName('id_product');        
        $IdProducts = $this->getAllMetaValues( $metakey );        
        if ( $IdProducts && is_array($IdProducts) ) {
                
            foreach($IdProducts as $IdProduct) {
                $response = $this->remote_delete_product($IdProduct);                
                usleep( (0.02 * 1000000) );
            }
        }
    }    

    public static function remove_product_metadata($product_id=0, $remove_all=false) {
        $flag_remove_all = ($product_id === 0 && $remove_all) ? true : false;

        $object_type = 'post';        
        // custom metadata ws del prodotto
        //$params=$this->main()->mapper->getWorthstockCustomParamList();
        $params = WorthstockDataMapper::getWorthstockCustomParamList();
        
        foreach($params as $value) {
            delete_metadata($object_type, $product_id, WorthstockDataMapper::getWorthstockCustomParamName($value), null, $flag_remove_all);
        }

        // metadata del prodotto serializzato
        $metakey = '_wst_temp_serialized_product';
        delete_metadata($object_type, $product_id, $metakey, null, $flag_remove_all);

        // product last operation        
        $metakey = '_worthstock_last_operation';
        delete_metadata($object_type, $product_id, $metakey, null, $flag_remove_all);
        
        $metakey = WorthstockDataMapper::getWorthstockCustomParamName('id_product');
        delete_metadata($object_type, $product_id, $metakey, null, false);

    }

    /**
     * FROM HERE ON CODE TO UPDATE DATA ON THE REMOTE SIDE
     */

    // NOTIFICATIONS HANDLERS
    
    function notify_stock_updated( $product_id ) {
        if ($this->main()->get_woocommerce_replication() !== true) return false;
        // get an instance of the WC_Product Object
        //$product      = wc_get_product( $product_id );
        
        $data = $this->main()->mapper->convertProductToWorthstock($product_id, 'quantity');

        if (count($data)) {
            foreach($data as $product) {
                $response = $this->remote_update_quantity($product);
            }
        }
        
        if ( $this->check_response_status($response) === true) {       
            // salvo per futuri controlli
            $this->main()->queue->save_serialized_product($product_id);
        }

        $this->log_status_operations('Notifica quantità', $response, $product_id);
        return $response;

        /* $stock_qty    = $product->get_stock_quantity();
        $event_id     = $product->get_attribute('event_id');
        $product_name = $product->get_name(); */
    }
    
    function notify_price_updated( $product_id ) {
        if ($this->main()->get_woocommerce_replication() !== true) return false;
        // get an instance of the WC_Product Object
        $data = $this->main()->mapper->convertProductToWorthstock($product_id, 'price');
        
        if (count($data)) {
            foreach($data as $product) {
                $response = $this->remote_update_prices($product);
            }
        }
        
        if ( $this->check_response_status($response) === true) {       
            // salvo per futuri controlli
            $this->main()->queue->save_serialized_product($product_id);
        }

        $this->log_status_operations('Notifica prezzo', $response, $product_id);
        return $response;
    }

    function notify_product_inserted( $product_id ) {
        // TODO: disabilitare sync se in draft?
        if ($this->main()->get_woocommerce_replication() !== true) return false;
        $data = $this->main()->mapper->convertProductToWorthstock($product_id, 'insert');
        // if product doesn't exist
        if ($data === false) return false;
        $response = $this->remote_insert_product($data);
        if ( $this->check_response_status($response) === true) {
            // save worthstock metas
            $response_data = $response->results;
            $this->save_worthstock_custom_product_meta($product_id, $response_data, 'insert');
        }

        $this->log_status_operations('Notifica inserimento prodotto', $response, $product_id);        
        //if (WORTHSTOCK__DEBUG === true) {
        if ($this->main()->settings->get_option( 'worthstock_debug_flag' ) == 'yes' ) {
            $this->api_logger('REQUEST ' . $this->prepare_product_json('insert', $data));
        }
        return ['request' => $data, 'response' => $response];
        //return $response;
    }

    function notify_product_updated( $product_id ) {
        if ($this->main()->get_woocommerce_replication() !== true) return false;
        $data = $this->main()->mapper->convertProductToWorthstock($product_id, 'update');
        // if product doesn't exist
        if ($data === false) return false;

        $IdProducts = get_post_meta($product_id, $this->main()->mapper->getWorthstockCustomParamName('id_product'));
        
        if ( count($IdProducts) <> count($data) ) {
            $this->notify_product_deleted($product_id);
        }

        // if worthstock product id is not save in custom meta assume product is missing on remote site, then request an insert. This is because woocommerce hook for new product is not reliable.
        // TODO: find a better method
        if ( ! $data["IdProduct"] && ! $data[0]["IdProduct"] ) {
            return $this->notify_product_inserted( $product_id );
        }
        
        $response = $this->remote_update_product($data);
        if ( $this->check_response_status($response) === true) {
            // save worthstock metas
            $response_data = $response->results;
            $this->save_worthstock_custom_product_meta($product_id, $response_data);
        }

        $this->log_status_operations('Notifica aggiornamento prodotto', $response, $product_id);
        if ($this->main()->settings->get_option( 'worthstock_debug_flag' ) == 'yes' ) {
            $this->api_logger('REQUEST ' . $this->prepare_product_json('update', $data));
        }
        return ['request' => $data, 'response' => $response];
    }
    
    function notify_product_deleted( $post_id ) {
        if ($this->main()->get_woocommerce_replication() !== true) return false;
        
        if( get_post_type( $post_id ) != 'product' ) return; // this is important not only to distinguish from 'post', but also 'post_variation'

        //$data = $this->main()->mapper->convertProductToWorthstock($post_id, 'price');
        
        $IdProducts = get_post_meta($post_id, $this->main()->mapper->getWorthstockCustomParamName('id_product'));
        
        if (count($IdProducts)) {
            foreach($IdProducts as $IdProduct) {
                $response = $this->remote_delete_product($IdProduct);
            }
        }
        // rimuovo i metadata
        $this->remove_product_metadata($post_id);

        $this->log_status_operations('Notifica rimozione prodotto', $response, $post_id);
        return $response;
    }
    
    
    public function remote_delete_product($id_product) {
        $url=$this->getEndPoint("api_delete_product");
        $final_json = '{
            "IdProduct": "'.$id_product.'"
        }';
        return $this->deleteProductCurl($url, $final_json);
    }

    /**
     * Function for PriceUpdate
     */
    public function remote_update_prices($product) {
        /*
         * Struttura
         * array Products(
         *  IdProduct => variabile $id_product
         *  array Variants(
         *      array(
         *        "IdVariant" int
                  "PriceLST"  float
                  "PriceB2C"  float
                  "PricePROMO" float
                  "PriceOVER" float
                )
         *  )
         * )
         */
        $url = $this->getEndPoint("api_put_prices");
        $final_json='{"Products": [{"IdProduct": "'.$product['IdProduct'].'","Variants": [';
        $final_json_data = array();
        foreach ($product['Variants'] as $variant) {
            $final_json_data[]='{"IdVariant": "'.$variant["IdProductVariant"].'","PriceLST": "'.$variant["PriceLST"].'","PriceB2C": "'.$variant["PriceB2C"].'","PricePROMO": "'.$variant["PricePROMO"].'","PriceOVER": "'.$variant["PriceOVER"].'"}';
        }
        $final_json .= implode(",",$final_json_data);
        $final_json .= ' ]}]}';
        return $this->callPutApi($url, $final_json);

    }

    /**
     * Function for QuantityUpdate
     */
    public function remote_update_quantity($product) {
        /*
         * Struttura
         * array Products(
         *  IdProduct => variabile $id_product
         *  array Variants(
         *      array(
         *        "IdVariant" int
                  "QuantityAvailable" int
                 )
         *  )
         * )
         */
        $url = $this->getEndPoint("api_put_quantity");
        $final_json='{"Products": [{"IdProduct": "'.$product['IdProduct'].'","Variants": [';
        $final_json_data=array();
        foreach ($product['Variants'] as $variant) {
            $final_json_data[]='{"IdVariant": "'.$variant["IdProductVariant"].'","QuantityAvailable": "'.$variant["QuantityAvailable"].'"}';
        }
        $final_json.=implode(",",$final_json_data);
        $final_json.=' ]}]}';
        return $this->callPutApi($url, $final_json);
    }


    public function remote_update_product($product_data) {
        $url=$this->getEndPoint("api_put_product");
        /* if (count($product_data)) {
            foreach($product_data as $product) {
                $json_string = ($this->prepare_product_json('update', [$product]));
                $response = $this->callPutApi($url, $json_string);
                $statuses[] = $response->status;
                $results[] = $response->results[0];
            }
        }
        return (object) array('status' => $statuses[0], 'results' => $results);
        */
        $json_string = ($this->prepare_product_json('update', $product_data));
        return $this->callPutApi($url, $json_string);
    }

    public function remote_insert_product($product_data) {
        $url=$this->getEndPoint("api_post_product");
        $json_string = ($this->prepare_product_json('insert', $product_data));
        return $this->insertProductCurl($url, $json_string);
    }
    
    /**
     * wrap up product data in the correct json structure
     */
    public function prepare_product_json($action_type,$product_data,$pretty_print=false) {
        //$data = array( "Products" => [$product_data], 'DebugParam' => $debug);
        $data = array( "Products" => $product_data);
        if ( count($product_data) > 1 && $action_type == "insert") {
            $data["FullResponse"] = true;  
        }
        return ($pretty_print) ? json_encode($data, JSON_PRETTY_PRINT) : json_encode($data);
    }
    
    /**
     * CURL functions performing remote calls
     */

    private function curlCaller($url, $post=false, $pars=null,$token=false) {

        $curl = curl_init();

        curl_setopt_array($curl, array(
            CURLOPT_URL => $url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING => '',
            CURLOPT_MAXREDIRS => 10,
            CURLOPT_TIMEOUT => 0,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST => 'POST',
            //CURLOPT_SSL_VERIFYPEER => 0, // Skip SSL Verification
            //CURLOPT_CAINFO => '/var/www/cacert.pem',
            CURLOPT_POSTFIELDS => array('Username' => $pars['Username'],'Password' => $pars['Password'],"WebhookUrl"=>$pars['WebhookUrl']),
        ));

        $response = curl_exec($curl);
        if ( curl_errno($curl) ) {
            throw new Exception(curl_error($curl));
        }
        // Closing
        curl_close($curl);
        // Print the return data
        return(json_decode($response));
    }

    protected function getGetData($url, $token) {
        $token = $this->getToken(); // leggiamo token da funzione
        $curl = curl_init();
        $auth = 'Authorization: Bearer '.$token;
        curl_setopt_array($curl, array(
            CURLOPT_URL => $url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING => '',
            CURLOPT_MAXREDIRS => 10,
            CURLOPT_TIMEOUT => 0,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
            CURLOPT_SSL_VERIFYPEER => 0, // Skip SSL Verification
            CURLOPT_CUSTOMREQUEST => 'GET',
            CURLOPT_HTTPHEADER => array(
                $auth
            ),
        ));

        $response = curl_exec($curl);
        if ( curl_errno($curl) ) {
            throw new Exception(curl_error($curl));
        }
        curl_close($curl);
        return json_decode($response);
    }    

    protected function callPutApi($api_url, $final_json){
        $token = $this->getToken(); // leggiamo token da funzione
        /* $args = array(
            'headers'     => array(
                'Authorization' => 'Bearer ' . $token,
            ),
            'body' => json_decode($final_json),
            'method' => 'PUT',
        ); 
        $response = wp_remote_request( $api_url, $args );
        return $response; */

        $curl = curl_init();
        curl_setopt_array($curl, array(
            CURLOPT_URL => $api_url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING => '',
            CURLOPT_MAXREDIRS => 10,
            CURLOPT_TIMEOUT => 0,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST => 'PUT',
            CURLOPT_POSTFIELDS => $final_json,
            CURLOPT_HTTPHEADER => array(
                'Authorization: Bearer '.$token,
                'Content-Type: text/plain',
            ),
        ));
        $response = curl_exec($curl);
        if ( curl_errno($curl) ) {
            throw new Exception(curl_error($curl));
        }
        curl_close($curl);
        return json_decode($response);
    }

    protected function insertProductCurl($url,$data){
        $token = $this->getToken(); // leggiamo token da funzione
        //echo "<br>insert con data :<br> $data <br> a url:$url<br>";
        $curl = curl_init();
        curl_setopt_array($curl, array(
            CURLOPT_URL => $url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING => '',
            CURLOPT_MAXREDIRS => 10,
            CURLOPT_TIMEOUT => 0,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST => 'POST',
            CURLOPT_POSTFIELDS => $data,
            CURLOPT_HTTPHEADER => array(
                'Authorization: Bearer '.$token,
                'Content-Type: text/plain',
            ),
        ));
        $response = curl_exec($curl);
        if ( curl_errno($curl) ) {
            throw new Exception(curl_error($curl));
        }
        curl_close($curl);
        return json_decode($response);
    }

    public function deleteProductCurl($url, $final_json) {        
        $token = $this->getToken(); // leggiamo token da funzione
        $curl = curl_init();
        curl_setopt_array($curl, array(
            CURLOPT_URL => $url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING => '',
            CURLOPT_MAXREDIRS => 10,
            CURLOPT_TIMEOUT => 0,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST => 'DELETE',
            /* CURLOPT_POSTFIELDS =>'{
                "IdProduct": "'.$id_product.'"
            }',  */           
            CURLOPT_POSTFIELDS => $final_json,
            CURLOPT_HTTPHEADER => array(
                'Authorization: Bearer '.$token,
                'Content-Type: text/plain',),
        ));
        $response = curl_exec($curl);
        if ( curl_errno($curl) ) {
            throw new Exception(curl_error($curl));
        }
        curl_close($curl);
        return json_decode($response);
    }

}